from __future__ import annotations

import abc
import asyncio
import functools
import json
import logging
import os
import random
import socket
import tempfile
import threading
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from typing import TYPE_CHECKING, AsyncIterator

from xiaogpt.utils import get_hostname

if TYPE_CHECKING:
    from typing import TypeVar

    from miservice import MiNAService

    from xiaogpt.config import Config

    T = TypeVar("T", bound="TTS")

logger = logging.getLogger(__name__)


class TTS(abc.ABC):
    """An abstract base class for Text-to-Speech models."""

    def __init__(
        self, mina_service: MiNAService, device_id: str, config: Config
    ) -> None:
        self.mina_service = mina_service
        self.device_id = device_id
        self.config = config

    async def wait_for_duration(self, duration: float) -> None:
        """Wait for the specified duration."""
        await asyncio.sleep(duration)
        while True:
            if not await self.get_if_xiaoai_is_playing():
                break
            await asyncio.sleep(1)

    async def get_if_xiaoai_is_playing(self):
        playing_info = await self.mina_service.player_get_status(self.device_id)
        # WTF xiaomi api
        is_playing = (
            json.loads(playing_info.get("data", {}).get("info", "{}")).get("status", -1)
            == 1
        )
        return is_playing

    @abc.abstractmethod
    async def synthesize(self, lang: str, text_stream: AsyncIterator[str]) -> None:
        """Synthesize speech from a stream of text."""
        raise NotImplementedError


class HTTPRequestHandler(SimpleHTTPRequestHandler):
    def log_message(self, format, *args):
        logger.debug(f"{self.address_string()} - {format}", *args)

    def log_error(self, format, *args):
        logger.error(f"{self.address_string()} - {format}", *args)

    def copyfile(self, source, outputfile):
        try:
            super().copyfile(source, outputfile)
        except (socket.error, ConnectionResetError, BrokenPipeError):
            # ignore this or TODO find out why the error later
            pass


class AudioFileTTS(TTS):
    """A TTS model that generates audio files locally and plays them via URL."""

    def __init__(
        self, mina_service: MiNAService, device_id: str, config: Config
    ) -> None:
        super().__init__(mina_service, device_id, config)
        self.dirname = tempfile.TemporaryDirectory(prefix="xiaogpt-tts-")
        self._start_http_server()

    @abc.abstractmethod
    async def make_audio_file(self, lang: str, text: str) -> tuple[Path, float]:
        """Synthesize speech from text and save it to a file.
        Return the file path and the duration of the audio in seconds.
        The file path must be relative to the self.dirname.
        """
        raise NotImplementedError

    async def synthesize(self, lang: str, text_stream: AsyncIterator[str]) -> None:
        queue: asyncio.Queue[tuple[str, float]] = asyncio.Queue()
        finished = asyncio.Event()

        async def worker():
            async for text in text_stream:
                path, duration = await self.make_audio_file(lang, text)
                url = f"http://{self.hostname}:{self.port}/{path.name}"
                await queue.put((url, duration))
            finished.set()

        task = asyncio.create_task(worker())

        while True:
            try:
                url, duration = queue.get_nowait()
            except asyncio.QueueEmpty:
                if finished.is_set():
                    break
                else:
                    await asyncio.sleep(0.1)
                    continue
            logger.debug("Playing URL %s (%s seconds)", url, duration)
            await asyncio.gather(
                self.mina_service.play_by_url(self.device_id, url),
                self.wait_for_duration(duration),
            )
        await task

    def _start_http_server(self):
        # set the port range
        port_range = range(8050, 8090)
        # get a random port from the range
        self.port = int(os.getenv("XIAOGPT_PORT", random.choice(port_range)))
        # create the server
        handler = functools.partial(HTTPRequestHandler, directory=self.dirname.name)
        httpd = ThreadingHTTPServer(("", self.port), handler)
        # start the server in a new thread
        server_thread = threading.Thread(target=httpd.serve_forever)
        server_thread.daemon = True
        server_thread.start()

        self.hostname = get_hostname()
        logger.info(f"Serving on {self.hostname}:{self.port}")



import httpx
import openai

from xiaogpt.bot.chatgptapi_bot import ChatGPTBot


class YiBot(ChatGPTBot):
    name = "yi"
    default_options = {"model": "yi-34b-chat-0205"}

    def __init__(
        self, yi_api_key: str, api_base="https://api.lingyiwanwu.com/v1"
    ) -> None:
        self.yi_api_key = yi_api_key
        self.api_base = api_base
        self.history: list[tuple[str, str]] = []

    def _make_openai_client(self, sess: httpx.AsyncClient) -> openai.AsyncOpenAI:
        return openai.AsyncOpenAI(
            api_key=self.yi_api_key, http_client=sess, base_url=self.api_base
        )

    @classmethod
    def from_config(cls, config):
        return cls(
            yi_api_key=config.yi_api_key,
            api_base="https://api.lingyiwanwu.com/v1",
        )



"""ChatGLM bot"""

from __future__ import annotations

from typing import Any

from rich import print

from xiaogpt.bot.base_bot import BaseBot, ChatHistoryMixin


class GLMBot(ChatHistoryMixin, BaseBot):
    name = "Chat GLM"
    default_options = {"model": "chatglm_turbo"}

    def __init__(self, glm_key: str) -> None:
        from zhipuai import ZhipuAI

        self.model = "glm-4"  # Change glm model here

        self.history = []
        self.client = ZhipuAI(api_key=glm_key)

    @classmethod
    def from_config(cls, config):
        return cls(glm_key=config.glm_key)

    def ask(self, query, **options):
        ms = self.get_messages()
        kwargs = {**self.default_options, **options}
        kwargs["model"] = self.model
        ms.append({"role": "user", "content": f"{query}"})
        kwargs["messages"] = ms
        try:
            r = self.client.chat.completions.create(**kwargs)
        except Exception as e:
            print(str(e))
            return
        message = r.choices[0].message.content

        self.add_message(query, message)
        print(message)
        return message

    async def ask_stream(self, query: str, **options: Any):
        ms = self.get_messages()
        kwargs = {**self.default_options, **options}
        kwargs["model"] = self.model
        ms.append({"role": "user", "content": f"{query}"})
        kwargs["messages"] = ms
        kwargs["stream"] = True
        try:
            r = self.client.chat.completions.create(**kwargs)
        except Exception as e:
            print(str(e))
            return
        full_content = ""
        for chunk in r:
            content = chunk.choices[0].delta.content
            full_content += content
            print(content, end="")
            yield content
        self.add_message(query, full_content)



from __future__ import annotations

import asyncio
import os

from langchain.memory import ConversationBufferWindowMemory
from rich import print

from xiaogpt.bot.base_bot import BaseBot
from xiaogpt.langchain.callbacks import AsyncIteratorCallbackHandler
from xiaogpt.langchain.chain import agent_search
from xiaogpt.utils import split_sentences


class LangChainBot(BaseBot):
    name = "Lang Chain"

    def __init__(
        self,
        openai_key: str,
        serpapi_api_key: str,
        proxy: str | None = None,
        api_base: str | None = None,
    ) -> None:
        # Set environment indicators
        os.environ["OPENAI_API_KEY"] = openai_key
        os.environ["SERPAPI_API_KEY"] = serpapi_api_key
        if api_base:
            os.environ["OPENAI_API_BASE"] = api_base
        if proxy:
            os.environ["OPENAI_PROXY"] = proxy
        self.memory = ConversationBufferWindowMemory(return_messages=True)

    def has_history(self) -> bool:
        return len(self.memory.chat_memory.messages) > 0

    def change_prompt(self, new_prompt: str) -> None:
        self.memory.clear()
        self.memory.chat_memory.add_user_message(new_prompt)

    @classmethod
    def from_config(cls, config):
        return cls(
            openai_key=config.openai_key,
            serpapi_api_key=config.serpapi_api_key,
            proxy=config.proxy,
            api_base=config.api_base,
        )

    async def ask(self, query, **options):
        return await agent_search(query, self.memory)

    async def ask_stream(self, query, **options):
        callback = AsyncIteratorCallbackHandler()
        task = asyncio.create_task(agent_search(query, self.memory, callback))
        try:
            async for message in split_sentences(callback.aiter()):
                yield message
        except Exception as e:
            print("An error occurred:", str(e))
        finally:
            print()
            await task



from typing import AsyncIterator

from miservice import MiIOService, MiNAService, miio_command

from xiaogpt.config import Config
from xiaogpt.tts.base import TTS
from xiaogpt.utils import calculate_tts_elapse


class MiTTS(TTS):
    def __init__(
        self, mina_service: MiNAService, device_id: str, config: Config
    ) -> None:
        super().__init__(mina_service, device_id, config)
        self.miio_service = MiIOService(mina_service.account)

    async def say(self, text: str) -> None:
        if not self.config.use_command:
            try:
                await self.mina_service.text_to_speech(self.device_id, text)
            except Exception:
                pass
        else:
            await miio_command(
                self.miio_service,
                self.config.mi_did,
                f"{self.config.tts_command} {text}",
            )

    async def synthesize(self, lang: str, text_stream: AsyncIterator[str]) -> None:
        async for text in text_stream:
            await self.say(text)
            await self.wait_for_duration(calculate_tts_elapse(text))



import email
import html
import imaplib
import re
import smtplib
from datetime import datetime, timedelta
from email.mime.text import MIMEText

from bs4 import BeautifulSoup


class Mailbox:
    # Gmail account settings need to be configured
    gmail_address = ""
    gmail_password = ""

    # Connect to IMAP server
    imap_server = "imap.gmail.com"
    imap_port = 993

    # Define email recipients and support adding multiple required email addresses
    to_addresses = [""]

    # Define the number of emails to summarize
    max_emails = 3

    def get_all_work_summary(self):
        print("Getting mail...")
        try:
            # Establish an IMAP connection
            mailbox = imaplib.IMAP4_SSL(self.imap_server, self.imap_port)
            # Log in to your Gmail account
            mailbox.login(self.gmail_address, self.gmail_password)
            # Select email
            mailbox.select("INBOX")
            # Get today's date
            today = datetime.now().strftime("%d-%b-%Y")
            # Build search criteria
            search_criteria = f'(SINCE "{today}")'
            # Search for matching messages
            status, email_ids = mailbox.search(None, search_criteria)

            if status == "OK":
                email_ids = email_ids[0].split()
                print(f"Number of emails received today: {len(email_ids)}")
                # Limit fetching up to max_emails emails
                max_emails = min(len(email_ids), self.max_emails)
                all_email_content = ""

                for i in range(max_emails):
                    email_id = email_ids[i]
                    email_content = self.get_email_content(mailbox, email_id)
                    if email_content:
                        all_email_content += f"{i+1}、{email_content}\n"

                # print(all_email_content)

            # close connection
            mailbox.logout()

            return all_email_content
        except Exception as e:
            print("Failed to get email:", str(e))

    def get_email_content(self, mailbox, email_id):
        # Get email content
        status, email_data = mailbox.fetch(email_id, "(RFC822)")
        if status == "OK":
            raw_email = email_data[0][1]
            msg = email.message_from_bytes(raw_email)

            # Get sender
            sender = msg["From"]
            # Extract senders within angle brackets (<>)
            sender = re.findall(r"<(.*?)>", sender)
            sender = sender[0] if sender else ""

            # Check whether the sender's email address ends with '.com', expand the field, and use it for email sender filtering
            if sender.lower().endswith(".com") and not msg["In-Reply-To"]:
                # Get email content
                email_content = ""
                if msg.is_multipart():
                    for part in msg.walk():
                        content_type = part.get_content_type()
                        if content_type == "text/plain":
                            email_content = part.get_payload(decode=True).decode(
                                "utf-8"
                            )
                            break
                        elif content_type == "text/html":
                            email_content = part.get_payload(decode=True).decode(
                                "utf-8"
                            )
                            email_content = html.unescape(
                                email_content
                            )  # Filter HTML code
                            break
                else:
                    email_content = msg.get_payload(decode=True).decode("utf-8")

                # Use BeautifulSoup to filter the html code if it still contains it
                if "html" in email_content.lower():
                    soup = BeautifulSoup(email_content, "html.parser")
                    email_content = soup.get_text()

                # Output text format
                email_content = re.sub(r"\s+", "", email_content)
                # Filter content between = signs
                email_content = re.sub(r"=\?.*?\?=", "", email_content)
                # Filter --content after the symbol
                email_content = re.sub(r"---.*", "", email_content)

                return f"{sender}Send an email with the content{email_content}"

        return ""

    def get_summary_by_ai(self, email_content: str, prompt: str) -> str:
        import openai

        print("Asking AI to summarize email content...")

        # Request ChatGPT for summary
        response = openai.chat.completions.create(
            model="gpt-3.5-turbo-0613",
            messages=[
                {"role": "system", "content": prompt},
                {"role": "user", "content": email_content},
            ],
        )

        # Extract summary generated by ChatGPT
        summary = response.choices[0].message.content.strip()
        # print(summary)
        return summary

    def send_mail(self, summary, theme="Email summary summary"):
        # Set senders and recipients
        from_address = self.gmail_address
        to_addresses = self.to_addresses  # Add multiple recipient email addresses

        # Build email content
        yesterday = (datetime.now() - timedelta(days=0)).strftime("%Y-%m-%d")
        subject = yesterday + theme
        body = summary

        try:
            # Connect to SMTP server
            smtp_server = smtplib.SMTP("smtp.gmail.com", 587)
            smtp_server.ehlo()
            smtp_server.starttls()
            # Login E-mail
            smtp_server.login(self.gmail_address, self.gmail_password)

            for to_address in to_addresses:
                # Create a plain text mail message object
                message = MIMEText(body, "plain", "utf-8")
                message["Subject"] = subject
                message["From"] = from_address
                message["To"] = to_address

                # send email
                smtp_server.sendmail(from_address, to_address, message.as_string())
                print("Email sent successfully to:", to_address)

            # close connection
            smtp_server.quit()
            print("All emails have been sent successfully!")
            return True
        except Exception as e:
            print("Email sending failed:", str(e))



from __future__ import annotations

import asyncio
from typing import Any, AsyncIterator
from uuid import UUID

from langchain.callbacks.base import AsyncCallbackHandler


class AsyncIteratorCallbackHandler(AsyncCallbackHandler):
    """Callback handler that returns an async iterator."""

    @property
    def always_verbose(self) -> bool:
        return True

    def __init__(self) -> None:
        self.queue = asyncio.Queue()
        self.done = asyncio.Event()

    async def on_chain_start(
        self,
        serialized: dict[str, Any],
        inputs: dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: list[str] | None = None,
        metadata: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        self.done.clear()

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        if token is not None and token != "":
            print(token, end="", flush=True)
            self.queue.put_nowait(token)

    async def on_chain_end(
        self,
        outputs: dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        self.done.set()

    async def on_chain_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        self.done.set()

    async def aiter(self) -> AsyncIterator[str]:
        while not self.queue.empty() or not self.done.is_set():
            # Wait for the next token in the queue,
            # but stop waiting if the done event is set
            done, other = await asyncio.wait(
                [
                    # NOTE: If you add other tasks here, update the code below,
                    # which assumes each set has exactly one task each
                    asyncio.ensure_future(self.queue.get()),
                    asyncio.ensure_future(self.done.wait()),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            # Cancel the other task
            if other:
                other.pop().cancel()

            # Extract the value of the first completed task
            token_or_done = done.pop().result()

            # If the extracted value is the boolean True, the done event was set
            if token_or_done is True:
                break

            # Otherwise, the extracted value is a token, which we yield
            yield token_or_done



import httpx
from groq import Groq as openai
from groq import AsyncGroq as AsyncOpenAI

from xiaogpt.bot.chatgptapi_bot import ChatGPTBot


class LlamaBot(ChatGPTBot):
    name = "llama"
    default_options = {"model": "llama3-70b-8192"}

    def __init__(self, llama_api_key: str) -> None:
        self.llama_api_key = llama_api_key
        self.history: list[tuple[str, str]] = []

    def _make_openai_client(self, sess: httpx.AsyncClient) -> AsyncOpenAI:
        return AsyncOpenAI(
            api_key=self.llama_api_key, http_client=sess, base_url=self.api_base
        )

    @classmethod
    def from_config(cls, config):
        return cls(
            llama_api_key=config.llama_api_key,
        )



"""Google Gemini bot"""

from __future__ import annotations

from typing import Any

from rich import print

from xiaogpt.bot.base_bot import BaseBot, ChatHistoryMixin

generation_config = {
    "temperature": 0.7,
    "top_p": 1,
    "top_k": 1,
    "max_output_tokens": 4096,
}

safety_settings = [
    {"category": "HARM_CATEGORY_HARASSMENT", "threshold": "BLOCK_MEDIUM_AND_ABOVE"},
    {"category": "HARM_CATEGORY_HATE_SPEECH", "threshold": "BLOCK_MEDIUM_AND_ABOVE"},
    {
        "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
        "threshold": "BLOCK_MEDIUM_AND_ABOVE",
    },
    {
        "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
        "threshold": "BLOCK_MEDIUM_AND_ABOVE",
    },
]


class GeminiBot(ChatHistoryMixin, BaseBot):
    name = "Gemini"

    def __init__(self, gemini_key: str, gemini_api_domain: str) -> None:
        import google.generativeai as genai

        from google.auth import api_key

        credentials = api_key.Credentials(gemini_key)
        if len(gemini_api_domain) > 0:
            print("Use custom gemini_api_domain: " + gemini_api_domain)
            credentials._universe_domain = gemini_api_domain
            genai.configure(
                transport="rest",
                credentials=credentials,
                client_options={
                    "api_endpoint": "https://" + gemini_api_domain,
                    "universe_domain": gemini_api_domain,
                },
            )
        else:
            genai.configure(api_key=gemini_key)

        self.history = []
        model = genai.GenerativeModel(
            model_name="gemini-pro",
            generation_config=generation_config,
            safety_settings=safety_settings,
        )
        self.convo = model.start_chat()

    @classmethod
    def from_config(cls, config):
        return cls(
            gemini_key=config.gemini_key, gemini_api_domain=config.gemini_api_domain
        )

    async def ask(self, query, **options):
        response = self.convo.send_message(query)
        message = response.text.strip()
        print(message)
        if len(self.convo.history) > 10:
            self.convo.history = self.convo.history[2:]
        return message

    async def ask_stream(self, query: str, **options: Any):
        if len(self.convo.history) > 10:
            self.convo.history = self.convo.history[2:]
        response = self.convo.send_message(query, stream=True)
        for chunk in response:
            print(chunk.text)
            yield chunk.text



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, ClassVar

import httpx
from rich import print

from xiaogpt.bot.base_bot import BaseBot, ChatHistoryMixin
from xiaogpt.utils import split_sentences

if TYPE_CHECKING:
    import openai


@dataclasses.dataclass
class ChatGPTBot(ChatHistoryMixin, BaseBot):
    name: ClassVar[str] = "ChatGPT"
    default_options: ClassVar[dict[str, str]] = {"model": "gpt-4o-2024-05-13"}
    openai_key: str
    api_base: str | None = None
    proxy: str | None = None
    deployment_id: str | None = None
    history: list[tuple[str, str]] = dataclasses.field(default_factory=list, init=False)

    def _make_openai_client(self, sess: httpx.AsyncClient) -> openai.AsyncOpenAI:
        import openai

        if self.api_base and self.api_base.rstrip("/").endswith("openai.azure.com"):
            return openai.AsyncAzureOpenAI(
                api_key=self.openai_key,
                azure_endpoint=self.api_base,
                api_version="2024-02-15-preview",
                azure_deployment=self.deployment_id,
                http_client=sess,
            )
        else:
            return openai.AsyncOpenAI(
                api_key=self.openai_key, http_client=sess, base_url=self.api_base
            )

    @classmethod
    def from_config(cls, config):
        return cls(
            openai_key=config.openai_key,
            api_base=config.api_base,
            proxy=config.proxy,
            deployment_id=config.deployment_id,
        )

    async def ask(self, query, **options):
        ms = self.get_messages()
        ms.append({"role": "user", "content": f"{query}"})
        kwargs = {**self.default_options, **options}
        httpx_kwargs = {}
        if self.proxy:
            httpx_kwargs["proxies"] = self.proxy
        async with httpx.AsyncClient(trust_env=True, **httpx_kwargs) as sess:
            client = self._make_openai_client(sess)
            try:
                completion = await client.chat.completions.create(messages=ms, **kwargs)
            except Exception as e:
                print(str(e))
                return ""

            message = completion.choices[0].message.content
            self.add_message(query, message)
            print(message)
            return message

    async def ask_stream(self, query, **options):
        ms = self.get_messages()
        ms.append({"role": "user", "content": f"{query}"})
        kwargs = {**self.default_options, **options}
        httpx_kwargs = {}
        if self.proxy:
            httpx_kwargs["proxies"] = self.proxy
        async with httpx.AsyncClient(trust_env=True, **httpx_kwargs) as sess:
            client = self._make_openai_client(sess)
            try:
                completion = await client.chat.completions.create(
                    messages=ms, stream=True, **kwargs
                )
            except Exception as e:
                print(str(e))
                return

            async def text_gen():
                async for event in completion:
                    if not event.choices:
                        continue
                    chunk_message = event.choices[0].delta
                    if chunk_message.content is None:
                        continue
                    print(chunk_message.content, end="")
                    yield chunk_message.content

            message = ""
            try:
                async for sentence in split_sentences(text_gen()):
                    message += sentence
                    yield sentence
            finally:
                print()
                self.add_message(query, message)



from __future__ import annotations

import tempfile
from pathlib import Path

from miservice import MiNAService

from xiaogpt.config import Config
from xiaogpt.tts.base import AudioFileTTS


class TetosTTS(AudioFileTTS):
    def __init__(
        self, mina_service: MiNAService, device_id: str, config: Config
    ) -> None:
        from tetos import get_speaker

        super().__init__(mina_service, device_id, config)
        assert config.tts and config.tts != "mi"
        speaker_cls = get_speaker(config.tts)
        try:
            self.speaker = speaker_cls(**config.tts_options)
        except TypeError as e:
            raise ValueError(f"{e}. Please add them via `tts_options` config") from e

    async def make_audio_file(self, lang: str, text: str) -> tuple[Path, float]:
        output_file = tempfile.NamedTemporaryFile(
            suffix=".mp3", mode="wb", delete=False, dir=self.dirname.name
        )
        duration = await self.speaker.synthesize(text, output_file.name, lang=lang)
        return Path(output_file.name), duration



from __future__ import annotations

import argparse
import json
import os
from dataclasses import dataclass, field
from typing import Any, Iterable, Literal

import yaml

from xiaogpt.utils import validate_proxy

LATEST_ASK_API = "https://userprofile.mina.mi.com/device_profile/v2/conversation?source=dialogu&hardware={hardware}&timestamp={timestamp}&limit=2"
COOKIE_TEMPLATE = "deviceId={device_id}; serviceToken={service_token}; userId={user_id}"
WAKEUP_KEYWORD = "小爱同学"

HARDWARE_COMMAND_DICT = {
    # hardware: (tts_command, wakeup_command)
    "LX06": ("5-1", "5-5"),
    "L05B": ("5-3", "5-4"),
    "S12": ("5-1", "5-5"),  # 第一代小爱，型号MDZ-25-DA
    "S12A": ("5-1", "5-5"),
    "LX01": ("5-1", "5-5"),
    "L06A": ("5-1", "5-5"),
    "LX04": ("5-1", "5-4"),
    "L05C": ("5-3", "5-4"),
    "L17A": ("7-3", "7-4"),
    "X08E": ("7-3", "7-4"),
    "LX05A": ("5-1", "5-5"),  # 小爱红外版
    "LX5A": ("5-1", "5-5"),  # 小爱红外版
    "L07A": ("5-1", "5-5"),  # Redmi小爱音箱Play(l7a)
    "L15A": ("7-3", "7-4"),
    "X6A": ("7-3", "7-4"),  # 小米智能家庭屏6
    "X10A": ("7-3", "7-4"),  # 小米智能家庭屏10
    # add more here
}

DEFAULT_COMMAND = ("5-1", "5-5")

KEY_WORD = ("帮我", "请")
CHANGE_PROMPT_KEY_WORD = ("更改提示词",)
PROMPT = "以下请用300字以内回答，请只回答文字不要带链接"
# simulate_xiaoai_question
MI_ASK_SIMULATE_DATA = {
    "code": 0,
    "message": "Success",
    "data": '{"bitSet":[0,1,1],"records":[{"bitSet":[0,1,1,1,1],"answers":[{"bitSet":[0,1,1,1],"type":"TTS","tts":{"bitSet":[0,1],"text":"Fake Answer"}}],"time":1677851434593,"query":"Fake Question","requestId":"fada34f8fa0c3f408ee6761ec7391d85"}],"nextEndTime":1677849207387}',
}


@dataclass
class Config:
    hardware: str = "LX06"
    account: str = os.getenv("MI_USER", "")
    password: str = os.getenv("MI_PASS", "")
    openai_key: str = os.getenv("OPENAI_API_KEY", "")
    moonshot_api_key: str = os.getenv("MOONSHOT_API_KEY", "")
    yi_api_key: str = os.getenv("YI_API_KEY", "")
    llama_api_key: str = os.getenv("GROQ_API_KEY", "")  # use groq
    glm_key: str = os.getenv("CHATGLM_KEY", "")
    gemini_key: str = os.getenv("GEMINI_KEY", "")  # keep the old rule
    qwen_key: str = os.getenv("DASHSCOPE_API_KEY", "")  # keep the old rule
    serpapi_api_key: str = os.getenv("SERPAPI_API_KEY", "")
    gemini_api_domain: str = os.getenv(
        "GEMINI_API_DOMAIN", ""
    )  # 自行部署的 Google Gemini 代理
    volc_access_key: str = os.getenv("VOLC_ACCESS_KEY", "")
    volc_secret_key: str = os.getenv("VOLC_SECRET_KEY", "")
    proxy: str | None = None
    mi_did: str = os.getenv("MI_DID", "")
    keyword: Iterable[str] = KEY_WORD
    change_prompt_keyword: Iterable[str] = CHANGE_PROMPT_KEY_WORD
    prompt: str = PROMPT
    mute_xiaoai: bool = False
    bot: str = "chatgptapi"
    cookie: str = ""
    api_base: str | None = None
    deployment_id: str | None = None
    use_command: bool = False
    verbose: bool = False
    start_conversation: str = "开始持续对话"
    end_conversation: str = "结束持续对话"
    stream: bool = False
    tts: Literal[
        "mi", "edge", "azure", "openai", "baidu", "google", "volc", "minimax"
    ] = "mi"
    tts_options: dict[str, Any] = field(default_factory=dict)
    gpt_options: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        if self.proxy:
            validate_proxy(self.proxy)
        if (
            self.api_base
            and self.api_base.endswith(("openai.azure.com", "openai.azure.com/"))
            and not self.deployment_id
        ):
            raise Exception(
                "Using Azure OpenAI needs deployment_id, read this: "
                "https://learn.microsoft.com/en-us/azure/cognitive-services/openai/how-to/chatgpt?pivots=programming-language-chat-completions"
            )
        if self.bot in ["chatgptapi"]:
            if not self.openai_key:
                raise Exception(
                    "Using GPT api needs openai API key, please google how to"
                )

    @property
    def tts_command(self) -> str:
        return HARDWARE_COMMAND_DICT.get(self.hardware, DEFAULT_COMMAND)[0]

    @property
    def wakeup_command(self) -> str:
        return HARDWARE_COMMAND_DICT.get(self.hardware, DEFAULT_COMMAND)[1]

    @classmethod
    def from_options(cls, options: argparse.Namespace) -> Config:
        config = {}
        if options.config:
            config = cls.read_from_file(options.config)
        for key, value in vars(options).items():
            if value is not None and key in cls.__dataclass_fields__:
                config[key] = value
        if config.get("tts") == "volc":
            config.setdefault("tts_options", {}).setdefault(
                "access_key", config.get("volc_access_key")
            )
            config.setdefault("tts_options", {}).setdefault(
                "secret_key", config.get("volc_secret_key")
            )
        return cls(**config)

    @classmethod
    def read_from_file(cls, config_path: str) -> dict:
        result = {}
        with open(config_path, "rb") as f:
            if config_path.endswith(".json"):
                config = json.load(f)
            else:
                config = yaml.safe_load(f)
            for key, value in config.items():
                if value is None:
                    continue
                if key == "keyword":
                    if not isinstance(value, list):
                        value = [value]
                    value = [kw for kw in value if kw]
                elif key == "use_chatgpt_api":
                    key, value = "bot", "chatgptapi"
                elif key == "use_newbing":
                    key, value = "bot", "newbing"
                elif key == "use_glm":
                    key, value = "bot", "glm"
                elif key == "use_gemini":
                    key, value = "bot", "gemini"
                elif key == "use_qwen":
                    key, value = "bot", "qwen"
                elif key == "use_doubao":
                    key, value = "bot", "doubao"
                elif key == "use_moonshot":
                    key, value = "bot", "moonshot"
                elif key == "use_yi":
                    key, value = "bot", "yi"
                elif key == "use_llama":
                    key, value = "bot", "llama"
                elif key == "use_langchain":
                    key, value = "bot", "langchain"
                elif key == "enable_edge_tts":
                    key, value = "tts", "edge"
                if key in cls.__dataclass_fields__:
                    result[key] = value
        return result



from __future__ import annotations

from langchain.agents import AgentType, Tool, initialize_agent
from langchain.callbacks.base import BaseCallbackHandler
from langchain.chains import LLMMathChain
from langchain.schema.memory import BaseMemory
from langchain_community.chat_models import ChatOpenAI
from langchain_community.utilities import SerpAPIWrapper


async def agent_search(
    query: str, memeory: BaseMemory, callback: BaseCallbackHandler | None = None
) -> str:
    llm = ChatOpenAI(
        streaming=True,
        temperature=0,
        model="gpt-3.5-turbo-0613",
    )

    # Initialization: search chain, mathematical calculation chain
    search = SerpAPIWrapper()
    llm_math_chain = LLMMathChain.from_llm(llm=llm, verbose=False)

    # Tool list: search, mathematical calculations
    tools = [
        Tool(
            name="Search",
            func=search.run,
            description="如果你不知道或不确定答案，可以使用这个搜索引擎检索答案",
        ),
        Tool(
            name="Calculator",
            func=llm_math_chain.run,
            description="在需要回答数学问题时非常有用",
        ),
    ]

    agent = initialize_agent(
        tools, llm, agent=AgentType.OPENAI_FUNCTIONS, verbose=False, memory=memeory
    )
    callbacks = [callback] if callback else None
    # query eg：'杭州亚运会中国队获得了多少枚金牌？' // '计算3的2次方'
    return await agent.arun(query, callbacks=callbacks)



from __future__ import annotations

from xiaogpt.bot.base_bot import BaseBot
from xiaogpt.bot.chatgptapi_bot import ChatGPTBot
from xiaogpt.bot.doubao_bot import DoubaoBot
from xiaogpt.bot.gemini_bot import GeminiBot
from xiaogpt.bot.glm_bot import GLMBot
from xiaogpt.bot.langchain_bot import LangChainBot
from xiaogpt.bot.llama_bot import LlamaBot
from xiaogpt.bot.moonshot_bot import MoonshotBot
from xiaogpt.bot.qwen_bot import QwenBot
from xiaogpt.bot.yi_bot import YiBot
from xiaogpt.config import Config

BOTS: dict[str, type[BaseBot]] = {
    "chatgptapi": ChatGPTBot,
    "glm": GLMBot,
    "gemini": GeminiBot,
    "qwen": QwenBot,
    "langchain": LangChainBot,
    "doubao": DoubaoBot,
    "moonshot": MoonshotBot,
    "yi": YiBot,
    "llama": LlamaBot,
}


def get_bot(config: Config) -> BaseBot:
    try:
        return BOTS[config.bot].from_config(config)
    except KeyError:
        raise ValueError(f"Unsupported bot {config.bot}, must be one of {list(BOTS)}")


__all__ = [
    "ChatGPTBot",
    "GLMBot",
    "GeminiBot",
    "MoonshotBot",
    "QwenBot",
    "get_bot",
    "LangChainBot",
    "DoubaoBot",
    "YiBot",
    "LlamaBot",
]



import argparse
import asyncio

from xiaogpt.config import Config
from xiaogpt.xiaogpt import MiGPT


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--hardware",
        dest="hardware",
        help="小爱 hardware",
    )
    parser.add_argument(
        "--account",
        dest="account",
        help="xiaomi account",
    )
    parser.add_argument(
        "--password",
        dest="password",
        help="xiaomi password",
    )
    parser.add_argument(
        "--openai_key",
        dest="openai_key",
        help="openai api key",
    )
    parser.add_argument(
        "--moonshot_api_key",
        dest="moonshot_api_key",
        help="Moonshot api key",
    )
    parser.add_argument(
        "--llama_api_key",
        dest="llama_api_key",
        help="llama(use groq) api key",
    )
    parser.add_argument(
        "--yi_api_key",
        dest="yi_api_key",
        help="01wanwu api key",
    )
    parser.add_argument(
        "--glm_key",
        dest="glm_key",
        help="chatglm api key",
    )
    parser.add_argument(
        "--gemini_key",
        dest="gemini_key",
        help="gemini api key",
    )
    parser.add_argument(
        "--gemini_api_domain",
        dest="gemini_api_domain",
        help="custom gemini api domain",
    )
    parser.add_argument(
        "--qwen_key",
        dest="qwen_key",
        help="Alibaba Qwen api key",
    )
    parser.add_argument(
        "--serpapi_api_key",
        dest="serpapi_api_key",
        help="serp api key see https://serpapi.com/",
    )
    parser.add_argument(
        "--proxy",
        dest="proxy",
        help="http proxy url like http://localhost:8080",
    )
    parser.add_argument(
        "--cookie",
        dest="cookie",
        help="xiaomi cookie",
    )
    parser.add_argument(
        "--stream",
        dest="stream",
        action="store_true",
        default=None,
        help="GPT stream mode",
    )
    parser.add_argument(
        "--use_command",
        dest="use_command",
        action="store_true",
        default=None,
        help="use command to tts",
    )
    parser.add_argument(
        "--mute_xiaoai",
        dest="mute_xiaoai",
        action="store_true",
        default=None,
        help="try to mute xiaoai answer",
    )
    parser.add_argument(
        "--volc_access_key", dest="volc_access_key", help="Volcengine access key"
    )
    parser.add_argument(
        "--volc_secret_key", dest="volc_secret_key", help="Volcengine secret key"
    )
    parser.add_argument(
        "--verbose",
        dest="verbose",
        action="store_true",
        default=None,
        help="show info",
    )
    parser.add_argument(
        "--tts",
        help="TTS provider",
        choices=["mi", "edge", "openai", "azure", "google", "baidu", "volc"],
    )
    bot_group = parser.add_mutually_exclusive_group()
    bot_group.add_argument(
        "--use_chatgpt_api",
        dest="bot",
        action="store_const",
        const="chatgptapi",
        help="if use openai chatgpt api",
    )
    bot_group.add_argument(
        "--use_moonshot_api",
        dest="bot",
        action="store_const",
        const="moonshot",
        help="if use moonshot api",
    )
    bot_group.add_argument(
        "--use_yi_api",
        dest="bot",
        action="store_const",
        const="yi",
        help="if use yi api",
    )
    bot_group.add_argument(
        "--use_langchain",
        dest="bot",
        action="store_const",
        const="langchain",
        help="if use langchain",
    )
    bot_group.add_argument(
        "--use_glm",
        dest="bot",
        action="store_const",
        const="glm",
        help="if use chatglm",
    )
    bot_group.add_argument(
        "--use_qwen",
        dest="bot",
        action="store_const",
        const="qwen",
        help="if use qwen",
    )
    bot_group.add_argument(
        "--use_gemini",
        dest="bot",
        action="store_const",
        const="gemini",
        help="if use gemini",
    )
    bot_group.add_argument(
        "--use_doubao",
        dest="bot",
        action="store_const",
        const="doubao",
        help="if use doubao",
    )
    bot_group.add_argument(
        "--use_llama",  # use groq
        dest="bot",
        action="store_const",
        const="llama",
        help="if use groq llama3",
    )
    parser.add_argument(
        "--bing_cookie_path",
        dest="bing_cookie_path",
        help="new bing cookies path if use new bing",
    )
    bot_group.add_argument(
        "--bot",
        dest="bot",
        help="bot type",
        choices=[
            "chatgptapi",
            "glm",
            "gemini",
            "langchain",
            "qwen",
            "doubao",
            "moonshot",
            "yi",
            "llama",
        ],
    )
    parser.add_argument(
        "--config",
        dest="config",
        help="config file path",
    )
    # args to change api_base
    parser.add_argument(
        "--api_base",
        dest="api_base",
        help="specify base url other than the OpenAI's official API address",
    )

    parser.add_argument(
        "--deployment_id",
        dest="deployment_id",
        help="specify deployment id, only used when api_base points to azure",
    )

    options = parser.parse_args()
    config = Config.from_options(options)

    async def main(config: Config) -> None:
        miboy = MiGPT(config)
        try:
            await miboy.run_forever()
        finally:
            await miboy.close()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(config))


if __name__ == "__main__":
    main()



from langchain.tools import BaseTool
from xiaogpt.langchain.examples.email.mail_box import Mailbox


class MailSummaryTool(BaseTool):
    name = "MailSumary"
    description = "当被问到总结邮件相关时，会触发这个工具，进行今日邮件总结和发送。当调用工具完毕，只需要回复总结成功或失败即可，立即结束本次回答"

    def get_mail_summary(self) -> str:
        """
        总结邮件：对邮箱内收到的邮件进行总结，并发送到指定邮箱
        """
        mailbox = Mailbox()
        all_email_content = mailbox.get_all_work_summary()
        prompt = """
        要求你作为一名总编辑。根据输入的多封邮件，对每封做简明扼要的摘要。要求如下：
        1、对每封邮件摘要总结，摘要总结字数在25字以内
        2、排版按照 发送人：xx 内容：xx （换一行）
        3、注意换行，要求全文美观简洁
        4、展示邮件内提到项目名，不用额外扩展讲项目内容和进度
        """
        gpt_content = mailbox.get_summary_by_ai(all_email_content, prompt)
        is_success = mailbox.send_mail(gpt_content)
        if is_success:
            return "总结邮件成功"
        else:
            return "总结邮件失败，请检查邮箱配置"

    def _run(self, query: str) -> str:
        return self.get_mail_summary()

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("MailSummaryTool does not support async")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator, TypeVar

from xiaogpt.config import Config

T = TypeVar("T", bound="BaseBot")


class BaseBot(ABC):
    name: str

    @abstractmethod
    async def ask(self, query: str, **options: Any) -> str:
        pass

    @abstractmethod
    async def ask_stream(self, query: str, **options: Any) -> AsyncGenerator[str, None]:
        pass

    @classmethod
    @abstractmethod
    def from_config(cls: type[T], config: Config) -> T:
        pass

    @abstractmethod
    def has_history(self) -> bool:
        pass

    @abstractmethod
    def change_prompt(self, new_prompt: str) -> None:
        pass


class ChatHistoryMixin:
    history: list[tuple[str, str]]

    def has_history(self) -> bool:
        return bool(self.history)

    def change_prompt(self, new_prompt: str) -> None:
        if self.history:
            print(self.history)
            self.history[0][0] = new_prompt

    def get_messages(self) -> list[dict]:
        ms = []
        for h in self.history:
            ms.append({"role": "user", "content": h[0]})
            ms.append({"role": "assistant", "content": h[1]})
        return ms

    def add_message(self, query: str, message: str) -> None:
        self.history.append([f"{query}", message])
        # only keep 5 history
        first_history = self.history.pop(0)
        self.history = [first_history] + self.history[-5:]



#!/usr/bin/env python3
from __future__ import annotations

import asyncio
import functools
import json
import logging
import re
import time
from pathlib import Path
from typing import AsyncIterator

from aiohttp import ClientSession, ClientTimeout
from miservice import MiAccount, MiIOService, MiNAService, miio_command
from rich import print
from rich.logging import RichHandler

from xiaogpt.bot import get_bot
from xiaogpt.config import (
    COOKIE_TEMPLATE,
    LATEST_ASK_API,
    MI_ASK_SIMULATE_DATA,
    WAKEUP_KEYWORD,
    Config,
)
from xiaogpt.tts import TTS, MiTTS, TetosTTS
from xiaogpt.utils import (
    parse_cookie_string,
)

EOF = object()


class MiGPT:
    def __init__(self, config: Config):
        self.config = config

        self.mi_token_home = Path.home() / ".mi.token"
        self.last_timestamp = int(time.time() * 1000)  # timestamp last call mi speaker
        self.cookie_jar = None
        self.device_id = ""
        self.parent_id = None
        self.mina_service = None
        self.miio_service = None
        self.in_conversation = False
        self.polling_event = asyncio.Event()
        self.last_record = asyncio.Queue(1)
        # setup logger
        self.log = logging.getLogger("xiaogpt")
        self.log.setLevel(logging.DEBUG if config.verbose else logging.INFO)
        self.log.addHandler(RichHandler())
        self.log.debug(config)
        self.mi_session = ClientSession()

    async def close(self):
        await self.mi_session.close()

    async def poll_latest_ask(self):
        async with ClientSession() as session:
            session._cookie_jar = self.cookie_jar
            while True:
                self.log.debug(
                    "Listening new message, timestamp: %s", self.last_timestamp
                )
                new_record = await self.get_latest_ask_from_xiaoai(session)
                start = time.perf_counter()
                self.log.debug(
                    "Polling_event, timestamp: %s %s", self.last_timestamp, new_record
                )
                await self.polling_event.wait()
                if (
                    self.config.mute_xiaoai
                    and new_record
                    and self.need_ask_gpt(new_record)
                ):
                    await self.stop_if_xiaoai_is_playing()
                if (d := time.perf_counter() - start) < 1:
                    # sleep to avoid too many request
                    self.log.debug("Sleep %f, timestamp: %s", d, self.last_timestamp)
                    # if you want force mute xiaoai, comment this line below.
                    await asyncio.sleep(1 - d)

    async def init_all_data(self):
        await self.login_miboy()
        await self._init_data_hardware()
        self.mi_session.cookie_jar.update_cookies(self.get_cookie())
        self.cookie_jar = self.mi_session.cookie_jar
        self.tts  # init tts

    async def login_miboy(self):
        account = MiAccount(
            self.mi_session,
            self.config.account,
            self.config.password,
            str(self.mi_token_home),
        )
        # Forced login to refresh to refresh token
        await account.login("micoapi")
        self.mina_service = MiNAService(account)
        self.miio_service = MiIOService(account)

    async def _init_data_hardware(self):
        if self.config.cookie:
            # if use cookie do not need init
            return
        hardware_data = await self.mina_service.device_list()
        # fix multi xiaoai problems we check did first
        # why we use this way to fix?
        # some videos and articles already in the Internet
        # we do not want to change old way, so we check if miotDID in `env` first
        # to set device id

        for h in hardware_data:
            if did := self.config.mi_did:
                if h.get("miotDID", "") == str(did):
                    self.device_id = h.get("deviceID")
                    break
                else:
                    continue
            if h.get("hardware", "") == self.config.hardware:
                self.device_id = h.get("deviceID")
                break
        else:
            raise Exception(
                f"we have no hardware: {self.config.hardware} please use `micli mina` to check"
            )
        if not self.config.mi_did:
            devices = await self.miio_service.device_list()
            try:
                self.config.mi_did = next(
                    d["did"]
                    for d in devices
                    if d["model"].endswith(self.config.hardware.lower())
                )
            except StopIteration:
                raise Exception(
                    f"cannot find did for hardware: {self.config.hardware} "
                    "please set it via MI_DID env"
                )

    def get_cookie(self):
        if self.config.cookie:
            cookie_jar = parse_cookie_string(self.config.cookie)
            # set attr from cookie fix #134
            cookie_dict = cookie_jar.get_dict()
            self.device_id = cookie_dict["deviceId"]
            return cookie_jar
        else:
            with open(self.mi_token_home) as f:
                user_data = json.loads(f.read())
            user_id = user_data.get("userId")
            service_token = user_data.get("micoapi")[1]
            cookie_string = COOKIE_TEMPLATE.format(
                device_id=self.device_id, service_token=service_token, user_id=user_id
            )
            return parse_cookie_string(cookie_string)

    @functools.cached_property
    def chatbot(self):
        return get_bot(self.config)

    async def simulate_xiaoai_question(self):
        data = MI_ASK_SIMULATE_DATA
        # Convert the data['data'] value from a string to a dictionary
        data_dict = json.loads(data["data"])
        # Get the first item in the records list
        record = data_dict["records"][0]
        # Replace the query and time values with user input
        record["query"] = input("Enter the new query: ")
        record["time"] = int(time.time() * 1000)
        # Convert the updated data_dict back to a string and update the data['data'] value
        data["data"] = json.dumps(data_dict)
        await asyncio.sleep(1)

        return data

    def need_ask_gpt(self, record):
        if not record:
            return False
        query = record.get("query", "")
        return (
            self.in_conversation
            and not query.startswith(WAKEUP_KEYWORD)
            or query.lower().startswith(tuple(w.lower() for w in self.config.keyword))
        )

    def need_change_prompt(self, record):
        query = record.get("query", "")
        return query.startswith(tuple(self.config.change_prompt_keyword))

    def _change_prompt(self, new_prompt):
        new_prompt = re.sub(
            rf"^({'|'.join(self.config.change_prompt_keyword)})", "", new_prompt
        )
        new_prompt = "以下都" + new_prompt
        print(f"Prompt from {self.config.prompt} change to {new_prompt}")
        self.config.prompt = new_prompt
        self.chatbot.change_prompt(new_prompt)

    async def get_latest_ask_from_xiaoai(self, session: ClientSession) -> dict | None:
        retries = 3
        for i in range(retries):
            try:
                timeout = ClientTimeout(total=15)
                r = await session.get(
                    LATEST_ASK_API.format(
                        hardware=self.config.hardware,
                        timestamp=str(int(time.time() * 1000)),
                    ),
                    timeout=timeout,
                )
            except Exception as e:
                self.log.warning(
                    "Execption when get latest ask from xiaoai: %s", str(e)
                )
                continue
            try:
                data = await r.json()
            except Exception:
                self.log.warning("get latest ask from xiaoai error, retry")
                if i == 1:
                    # tricky way to fix #282 #272 # if it is the third time we re init all data
                    print("Maybe outof date trying to re init it")
                    await self._retry()
            else:
                return self._get_last_query(data)
        return None

    async def _retry(self):
        await self.init_all_data()

    def _get_last_query(self, data: dict) -> dict | None:
        if d := data.get("data"):
            records = json.loads(d).get("records")
            if not records:
                return None
            last_record = records[0]
            timestamp = last_record.get("time")
            if timestamp > self.last_timestamp:
                try:
                    self.last_record.put_nowait(last_record)
                    self.last_timestamp = timestamp
                    return last_record
                except asyncio.QueueFull:
                    pass
        return None

    async def do_tts(self, value):
        if not self.config.use_command:
            try:
                await self.mina_service.text_to_speech(self.device_id, value)
            except Exception:
                pass
        else:
            await miio_command(
                self.miio_service,
                self.config.mi_did,
                f"{self.config.tts_command} {value}",
            )

    @functools.cached_property
    def tts(self) -> TTS:
        if self.config.tts == "mi":
            return MiTTS(self.mina_service, self.device_id, self.config)
        else:
            return TetosTTS(self.mina_service, self.device_id, self.config)

    async def wait_for_tts_finish(self):
        while True:
            if not await self.get_if_xiaoai_is_playing():
                break
            await asyncio.sleep(1)

    @staticmethod
    def _normalize(message: str) -> str:
        message = message.strip().replace(" ", "--")
        message = message.replace("\n", "，")
        message = message.replace('"', "，")
        return message

    async def ask_gpt(self, query: str) -> AsyncIterator[str]:
        if not self.config.stream:
            if self.config.bot == "glm":
                answer = self.chatbot.ask(query, **self.config.gpt_options)
            else:
                answer = await self.chatbot.ask(query, **self.config.gpt_options)
            message = self._normalize(answer) if answer else ""
            yield message
            return

        async def collect_stream(queue):
            async for message in self.chatbot.ask_stream(
                query, **self.config.gpt_options
            ):
                await queue.put(message)

        def done_callback(future):
            queue.put_nowait(EOF)
            if future.exception():
                self.log.error(future.exception())

        self.polling_event.set()
        queue = asyncio.Queue()
        is_eof = False
        task = asyncio.create_task(collect_stream(queue))
        task.add_done_callback(done_callback)
        while True:
            if is_eof or not self.last_record.empty():
                break
            message = await queue.get()
            if message is EOF:
                break
            while not queue.empty():
                if (next_msg := queue.get_nowait()) is EOF:
                    is_eof = True
                    break
                message += next_msg
            if message:
                yield self._normalize(message)
        self.polling_event.clear()
        task.cancel()

    async def get_if_xiaoai_is_playing(self):
        playing_info = await self.mina_service.player_get_status(self.device_id)
        # WTF xiaomi api
        is_playing = (
            json.loads(playing_info.get("data", {}).get("info", "{}")).get("status", -1)
            >= 1
        )
        return is_playing

    async def stop_if_xiaoai_is_playing(self):
        is_playing = await self.get_if_xiaoai_is_playing()
        if is_playing:
            # stop it
            await self.mina_service.player_stop(self.device_id)

    async def wakeup_xiaoai(self):
        return await miio_command(
            self.miio_service,
            self.config.mi_did,
            f"{self.config.wakeup_command} {WAKEUP_KEYWORD} 0",
        )

    async def run_forever(self):
        await self.init_all_data()
        task = asyncio.create_task(self.poll_latest_ask())
        assert task is not None  # to keep the reference to task, do not remove this
        print(
            f"Running xiaogpt now, 用[green]{'/'.join(self.config.keyword)}[/]开头来提问"
        )
        print(f"或用[green]{self.config.start_conversation}[/]开始持续对话")
        while True:
            self.polling_event.set()
            new_record = await self.last_record.get()
            self.polling_event.clear()  # stop polling when processing the question
            query = new_record.get("query", "").strip()
            if query == self.config.start_conversation:
                if not self.in_conversation:
                    print("开始对话")
                    self.in_conversation = True
                    await self.wakeup_xiaoai()
                await self.stop_if_xiaoai_is_playing()
                continue
            elif query == self.config.end_conversation:
                if self.in_conversation:
                    print("结束对话")
                    self.in_conversation = False
                await self.stop_if_xiaoai_is_playing()
                continue

            # we can change prompt
            if self.need_change_prompt(new_record):
                print(new_record)
                self._change_prompt(new_record.get("query", ""))

            if not self.need_ask_gpt(new_record):
                self.log.debug("No new xiao ai record")
                continue

            # drop key words
            query = re.sub(rf"^({'|'.join(self.config.keyword)})", "", query)
            # llama3 is not good at Chinese, so we need to add prompt in it.
            if self.config.bot == "llama":
                query = f"你是一个基于llama3 的智能助手，请你跟我对话时，一定使用中文，不要夹杂一些英文单词，甚至英语短语也不能随意使用，但类似于 llama3 这样的专属名词除外, 问题是：{query}"

            print("-" * 20)
            print("问题：" + query + "？")
            if not self.chatbot.has_history():
                query = f"{query}，{self.config.prompt}"
            # some model can not detect the language code, so we need to add it

            if self.config.tts != "mi":  # mi only say Chinese
                query += (
                    "，并用本段话的language code作为开头，用|分隔，如：en-US|你好……"
                )

            if self.config.mute_xiaoai:
                await self.stop_if_xiaoai_is_playing()
            else:
                # waiting for xiaoai speaker done
                await asyncio.sleep(8)
            await self.do_tts(f"正在问{self.chatbot.name}请耐心等待")
            try:
                print(
                    "以下是小爱的回答: ",
                    new_record.get("answers", [])[0].get("tts", {}).get("text"),
                )
            except IndexError:
                print("小爱没回")
            print(f"以下是 {self.chatbot.name} 的回答: ", end="")
            try:
                await self.speak(self.ask_gpt(query))
            except Exception as e:
                print(f"{self.chatbot.name} 回答出错 {str(e)}")
            else:
                print("回答完毕")
            if self.in_conversation:
                print(f"继续对话, 或用`{self.config.end_conversation}`结束对话")
                await self.wakeup_xiaoai()

    async def speak(self, text_stream: AsyncIterator[str]) -> None:
        text = await text_stream.__anext__()
        # See if the first part contains language code(e.g. en-US|Hello world)
        lang, _, first_chunk = text.rpartition("|")
        if len(lang) > 7:
            # It is not a legal language code, discard it
            lang, first_chunk = "", text

        lang = (
            matches[0]
            if (matches := re.findall(r"([a-z]{2}-[A-Z]{2})", lang))
            else "zh-CN"
        )

        async def gen():  # reconstruct the generator
            yield first_chunk
            async for text in text_stream:
                yield text

        await self.tts.synthesize(lang, gen())



"""Qwen bot"""

from __future__ import annotations

from http import HTTPStatus
from typing import Any

from rich import print

from xiaogpt.bot.base_bot import BaseBot, ChatHistoryMixin


class QwenBot(ChatHistoryMixin, BaseBot):
    name = "Qian Wen"

    def __init__(self, qwen_key: str) -> None:
        import dashscope
        from dashscope.api_entities.dashscope_response import Role

        self.history = []
        dashscope.api_key = qwen_key

    @classmethod
    def from_config(cls, config):
        return cls(qwen_key=config.qwen_key)

    async def ask(self, query, **options):
        from dashscope import Generation
        from dashscope.api_entities.dashscope_response import Role

        # from https://help.aliyun.com/zh/dashscope/developer-reference/api-details
        self.history.append({"role": Role.USER, "content": query})

        response = Generation.call(
            Generation.Models.qwen_turbo,
            messages=self.history,
            result_format="message",  # set the result to be "message" format.
        )
        if response.status_code == HTTPStatus.OK:
            # append result to messages.
            content = response.output.choices[0]["message"]["content"]
            self.history.append(
                {
                    "role": response.output.choices[0]["message"]["role"],
                    "content": content,
                }
            )
            # keep last five
            first_history = self.history.pop(0)
            self.history = [first_history] + self.history[-5:]
            print(content)
            return content
        else:
            print(
                "Request id: %s, Status code: %s, error code: %s, error message: %s"
                % (
                    response.request_id,
                    response.status_code,
                    response.code,
                    response.message,
                )
            )
            # we need to pop the wrong history
            print(f"Will pop the wrong question {query}")
            self.history.pop()
            return "没有返回"

    async def ask_stream(self, query: str, **options: Any):
        from dashscope import Generation
        from dashscope.api_entities.dashscope_response import Role

        self.history.append({"role": Role.USER, "content": query})
        responses = Generation.call(
            Generation.Models.qwen_turbo,
            messages=self.history,
            result_format="message",  # set the result to be "message" format.
            stream=True,
            incremental_output=True,  # get streaming output incrementally
        )
        full_content = ""  # with incrementally we need to merge output.
        role = None
        for response in responses:
            if response.status_code == HTTPStatus.OK:
                content = response.output.choices[0]["message"]["content"]
                full_content += content
                if not role:
                    role = response.output.choices[0]["message"]["role"]
                print(content, end="")
                yield content
            else:
                print(
                    "Request id: %s, Status code: %s, error code: %s, error message: %s"
                    % (
                        response.request_id,
                        response.status_code,
                        response.code,
                        response.message,
                    )
                )
        self.history.append({"role": role, "content": full_content})
        first_history = self.history.pop(0)
        self.history = [first_history] + self.history[-5:]



"""ChatGLM bot"""

from __future__ import annotations

import json
from typing import Any, AsyncIterator

import httpx
from rich import print

from xiaogpt.bot.base_bot import BaseBot, ChatHistoryMixin
from xiaogpt.config import Config
from xiaogpt.utils import split_sentences


class DoubaoBot(ChatHistoryMixin, BaseBot):
    API_URL = "https://maas-api.ml-platform-cn-beijing.volces.com"
    name = "豆包"
    default_options = {"model": "skylark-chat"}

    def __init__(self, access_key: str, secret_key: str) -> None:
        from tetos.volc import VolcSignAuth

        self.auth = VolcSignAuth(access_key, secret_key, "ml_maas", "cn-beijing")
        self.history = []

    @classmethod
    def from_config(cls, config: Config):
        return cls(access_key=config.volc_access_key, secret_key=config.volc_secret_key)

    def _get_data(self, query: str, **options: Any):
        options = {**self.default_options, **options}
        model = options.pop("model")
        ms = self.get_messages()
        ms.append({"role": "user", "content": query})
        return {"model": {"name": model}, "parameters": options, "messages": ms}

    async def ask(self, query, **options):
        data = self._get_data(query, **options)
        async with httpx.AsyncClient(base_url=self.API_URL, auth=self.auth) as client:
            resp = await client.post("/api/v1/chat", json=data)
            resp.raise_for_status()
            try:
                message = resp.json()["choice"]["message"]["content"]
            except Exception as e:
                print(str(e))
                return
            self.add_message(query, message)
            print(message)
            return message

    async def ask_stream(self, query: str, **options: Any):
        data = self._get_data(query, **options)
        data["stream"] = True

        async def sse_gen(line_iter: AsyncIterator[str]) -> AsyncIterator[str]:
            message = ""
            async for chunk in line_iter:
                if not chunk.startswith("data:"):
                    continue
                message = chunk[5:].strip()
                if message == "[DONE]":
                    break
                data = json.loads(message)
                text = data["choice"]["message"]["content"]
                print(text, end="", flush=True)
                message += text
                yield text
            print()
            self.add_message(query, message)

        async with httpx.AsyncClient(base_url=self.API_URL, auth=self.auth) as client:
            async with client.stream("POST", "/api/v1/chat", json=data) as resp:
                resp.raise_for_status()
                async for sentence in split_sentences(sse_gen(resp.aiter_lines())):
                    yield sentence
